import json

try:
    from urllib import urlencode
    from urllib2 import urlopen, Request, HTTPError
except ImportError:
    from urllib.error import HTTPError
    from urllib.request import urlopen, Request
    from urllib.parse import urlencode

import os

__author__ = 'Brandon Curtis'
__version__ = '1.0.1'
USER_AGENT = 'bukget %s' % __version__
BASE = 'http://api.bukget.org/3'


def _request(url, data=None, jsonify=True, headers={}, query={}):
    '''Base Request
    The parent function that all public functions call in order to initiate
    communication to the API.  Simply converts the Pythonic values into the
    the URL format expected by the API.  Also will add the User-Agent defined
    by the USER_AGENT variable.
    '''

    # There is no reason for there ever to be a callback specified.  If we see
    # one, just remove it.
    if 'callback' in query: del (query['callback'])

    # Here we will collapse the fields list if we see it into a string as is
    # expected by the API.
    if 'fields' in query and (' ' in query['fields'] or \
                                      isinstance(query['fields'], list)):
        query['fields'] = ','.join(query['fields'])
    if data is not None and 'fields' in data and (' ' in data['fields'] or \
                                                          isinstance(data['fields'], list)):
        data['fields'] = ','.join(data['fields'])

    # Append the query to the URL if we have anything in the query dictionary.
    if len(query) > 0:
        url += '?%s' % '&'.join(['%s=%s' % (q, query[q]) for q in query])

    # Lastly, if there is anything in the data variable, then urlencode it.
    if data is not None:
        data = urlencode(data).encode('utf-8')

    # Time to set the User-Agent string and actually query the API!
    headers['User-Agent'] = USER_AGENT
    try:
        response = urlopen(Request(BASE + url, data, headers)).read()
    except HTTPError:
        return None
    else:
        if jsonify:
            return json.loads(response.decode("utf-8"))
        else:
            return response


def plugins(server='', **query):
    '''Retreives a list of plugins.
    Optionally you can specity a server and any query variables per the API
    Documentation.
    '''
    call = '/plugins/%s' % server
    return _request(call, query=query)


def plugin_details(server, plugin, version='', **query):
    '''Retreives the plugin details.
    Optionally a specific version can be specified.  All query variables
    specified by the API docs will work here.
    '''
    call = '/plugins/%s/%s/%s' % (server, plugin, version)
    return _request(call, query=query)


def plugin_download(server, plugin, version):
    '''Downloads the plugin binary.
    This function will return the raw data stream from the API.  We will NOT be
    returning dictionaries in this case!
    '''
    call = '/plugins/%s/%s/%s/download' % (server, plugin, version)
    return _request(call, jsonify=False)


def authors():
    '''Returns a list of authors and their plugin counts from the API.'''
    return _request('/authors')


def author_plugins(author, server=None, **query):
    '''Retreives a list of plugins written by the specified author.
    Optionally you can specity a server and any query variables per the API
    Documentation.
    '''

    # Depending on if the server variable is set, we can have one of 2
    # different URLs, so here we will set the call to the correct one.
    if server is not None:
        call = '/authors/%s/%s' % (server, author)
    else:
        call = '/authors/%s' % author
    return _request(call, query=query)


def categories():
    '''Returns the category listing with the count of plugins for each cat.'''
    return _request('/categories')


def category_plugins(category, server=None, **query):
    '''Retreives a list of plugins in the specified category.
    Optionally you can specity a server and any query variables per the API
    Documentation.
    '''

    # Depending on if the server variable is set, we can have one of 2
    # different URLs, so here we will set the call to the correct one.
    if server is not None:
        call = '/categories/%s/%s' % (server, category)
    else:
        call = '/categories/%s' % category
    return _request(call, query=query)


def search(*filters, **query):
    '''Searching the API.
    This function is only utilizing the POST searching and is expecting properly
    formatted search dictionaries.  Also all of the same query variables as is
    described in the API3 docs will work here as well.
    '''
    query['filters'] = json.dumps(filters)
    return _request('/search', data=query)


def _levenshtein(s1, s2):
    """ Get the levenshtein edit distance between two strings
    """
    if len(s1) < len(s2):
        return _levenshtein(s2, s1)
    if len(s2) == 0:
        return len(s1)
    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[
                             j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row
    return previous_row[-1]


def _get_best_match(to_match, possible_matches):
    """ Return the value in possible_matches that has the smallest edit
    distance to to_match
    """
    lowest_distance = -1
    distances = {}
    for match in possible_matches:
        distance = _levenshtein(to_match, match)
        if lowest_distance < 0 or distance < lowest_distance:
            lowest_distance = distance
        if lowest_distance == 0:
            break
    for match in distances:
        if distances[match] == lowest_distance:
            return match
    return None


def find_by_name(server, name):
    ''' Find the slug by the name supplied
    First the name is turned all lowercase and checked if that is the slug.
    If it is, it's returned. Else it searches for plugins with equal names and
    chooses the first one. If no plugin has the name, it searches for a plugin
    with a smiliar name. Will return None if no slug was found
    '''
    # First we are testing if the name is the same as the slug
    try:
        if plugin_details(server, name.lower().replace(' ', '-'),
                          fields='slug') is not None:
            return name.lower().replace(' ', '-')
    except HTTPError:
        pass

    # Then we search for a plugin with name that matches
    search_result = search({
        'field': 'plugin_name',
        'action': '=', 'value': name
    }, {
        'field': 'server',
        'action': '=', 'value': server
    },
        fields='slug')
    if len(search_result) > 0:
        return search_result[0]['slug']

    # Then we search for a plugin with a name like it
    search_result = search({
        'field': 'plugin_name',
        'action': 'like', 'value': name
    }, {
        'field': 'server',
        'action': '=', 'value': server
    },
        fields='slug')
    if len(search_result) > 0:
        return _get_best_match(name, [i['slug'] for i in search_result])

    # No plugin found =(
    return None


def get_by_main(server, main):
    ''' 
    Fetches the slug of the plugin based on the java class name (main)
    '''
    search_result = search({
        'field': 'main',
        'action': '=', 'value': main
    }, {
        'field': 'server',
        'action': '=', 'value': server
    },
        fields='slug')
    if len(search_result) > 0:
        return search_result[0]['slug']
    else:
        return None


class BukkitResource(object):
    def __init__(self, json):
        self.json = json
        self.website = None
        self.bukkit_dev_page = None
        self.description = None
        self.logo = None
        self.versions = {}
        self.plugin_name = None
        self.popularity = {}
        self.server_type = None
        self.main_class = None
        self.authors = []
        self.slug = None
        self.categories = []
        self.stage = None
        self.latest_version = None
        self.__configure_from_json(json=json)

    def __configure_from_json(self, json):
        self.website = json['website']
        self.bukkit_dev_page = json['dbo_page']
        self.description = json['description']
        self.logo = json['logo']
        for version in json['versions']:
            self.versions[version['version']] = BukkitPluginVersion(json=version)

        self.plugin_name = json['plugin_name']
        self.popularity = json['popularity']
        self.server_type = json['server']
        self.main_class = json['main']
        self.authors = json['authors']
        self.slug = json['slug']
        self.categories = json['categories']
        self.stage = json['stage']
        self.latest_version = self.get_latest_version()

    def get_latest_version(self):
        if self.latest_version is None:
            self.latest_version = next(iter(self.versions.keys()))
        return self.latest_version

    def get_download_link(self, version="latest"):
        version_info = self.versions[self.latest_version if version == "latest" else version]
        return version_info.download

    def get_versioned_file_name(self, version="latest"):
        extension = os.path.splitext(self.get_download_link(version))[1]
        file_name = "%s-%s%s" % (self.plugin_name, self.latest_version if version == "latest" else version, extension)
        return file_name

    def get_command_info(self, command, version="latest"):
        # If version is latest get the command info for the latest plugin version, otherwise get it
        # from the version request
        if version == "latest":
            return self.versions[self.latest_version].commands[command]
        else:
            return self.versions[version].commands[command]

    def has_version(self, version):
        for plugin_version in self.versions:
            if plugin_version.version == version:
                return True

        return False

    @staticmethod
    def from_name(name, version='latest'):
        json = plugin_details('bukkit', name, version=version)
        if "error" in json:
            raise ValueError("Unable to retrieve plugin %s of version %s" % (name, version))
        return BukkitResource(json=json)

    def __str__(self):
        return "%s" % dict(
            website=self.website,
            bukkit_dev_page=self.bukkit_dev_page,
            description=self.description,
            logo=self.logo,
            versions=self.versions,
            latest_version=self.latest_version,
            plugin_name=self.plugin_name,
            popularity=self.popularity,
            server_type=self.server_type,
            main_class=self.main_class,
            authors=self.authors,
            slug=self.slug,
            categories=self.categories,
            stage=self.stage
        )


class BukkitPluginVersion(object):
    def __init__(self, json):
        self.json = json
        self.status = None
        self.commands = {}
        self.changelog = None
        self.filename = None
        self.supported_server_versions = []
        self.hard_dependencies = []
        self.date = None
        self.version = None
        self.link = None
        self.download = None
        self.md5 = None
        self.type = None
        self.slug = None
        self.soft_dependencies = []
        self.permissions = {}
        self.__configure_from_json(json=json)

    def __configure_from_json(self, json):
        self.status = json['status']
        # Loop through all the command json and create BukkitPluginCommand objects for each, indexed by their name.
        for command in json['commands']:
            self.commands[command['command']] = BukkitPluginCommand(json=command)

        self.changelog = json['changelog']
        self.filename = json['filename']
        # Create a list based on all the versions in the json 'game_versions'
        self.supported_server_versions = [version for version in json['game_versions']]
        self.hard_dependencies = json['hard_dependencies']
        self.date = json['date']
        self.version = json['version']
        self.link = json['link']
        self.download = json['download']
        self.md5 = json['md5']
        self.type = json['type']
        self.slug = json['slug']
        self.soft_dependencies = json['soft_dependencies']
        for permission in json['permissions']:
            self.permissions[permission['role']] = permission['default']

    def supports_version(self, game_version):
        for version in self.supported_server_versions:
            if game_version in version:
                return True

        return False

    def __str__(self):
        return "%s-%s: values{%s}" % (self.filename, self.version, dict(
            status=self.status,
            commands=[str(command) for command in self.commands],
            changelog=self.changelog,
            filename=self.filename,
            supported_server_versions=self.supported_server_versions,
            hard_dependencies=self.hard_dependencies,
            soft_dependencies=self.soft_dependencies,
            release_date=self.date,
            version=self.version,
            link=self.link,
            download=self.download,
            md5=self.md5,
            release_type=self.type,
            slug=self.slug,
            permissions=str(self.permissions)
        ))


class BukkitPluginCommand(object):
    def __init__(self, json):
        self.command = None
        self.usage = None
        self.aliases = []
        self.permission_message = None
        self.permission = None
        self.json = json
        self.__configure_from_jsn(json=json)

    def __configure_from_jsn(self, json):
        self.command = json['command']
        self.usage = json['usage']
        self.aliases = json['aliases']
        self.permission_message = json['permission-message']
        self.permission = json['permission']

    def __str__(self):
        return "%s" % dict(
            command=self.command,
            usage=self.usage,
            aliases=self.aliases,
            permission_message=self.permission_message,
            permission=self.permission
        )


class BukkitAuthor(object):
    def __init__(self, author):
        self.name = author
        self.plugins = {}
        self.__retrieve_author_plugins()

    def __retrieve_author_plugins(self):
        json = author_plugins(self.name, server="bukkit")
        for plugin_info in json:
            self.plugins[plugin_info['plugin_name']] = BukkitResource.from_name(plugin_info['plugin_name'])

    @staticmethod
    def from_name(name):
        return BukkitAuthor(author=name)
